/* ------------------------------------------------------------------------- */
/* Authors: Jean-Francois Giorgi                                             */
/*          giorgi@irisa.fr                                                  */
/* This code is Read8SVX.c from Electronic Arts IFF88                        */ 
/* modified for Smus2sco                                                     */
/* ------------------------------------------------------------------------- */

/* Read8SVX.c **********************************************************
 * 
 * Read a sound sample from an IFF file.  21Jan85 
 * 
 * By Steve Hayes, Electronic Arts. 
 * This software is in the public domain. 
 * 
 * Modified in July 90 for Smus2sco by J.F. Giorgi 
 * ********************************************************************** 
 */
 
/* ************************************************************************* */
#include "exec/types.h" 
#include "iff/gio.h" 
#include "iff/8SVX.h" 
#include "math.h"
/* ************************************************************************* */
/* exported and imported variables */
#define MAX_OCT 10
#define MOT2 (MAX_OCT*2)
ULONG oneshot[MAX_OCT],  /* # samples in the octave 1-shot part */
      repeat[MAX_OCT];   /* # samples in the octave repeat part */
UBYTE ctoct;       /* # of octaves of waveforms */

extern char *IPath; /* where are the instruments */
extern char *calloc();
LOCAL FILE *orchfile;
/* ************************************************************************* */
char *BaseName(path)
char *path;
{
  register char *d,*s;

  if (path==NULL) return(path);
  s = path;
  d = path;
  while(*s != (char)0) if (*s++ == '/') d=s;
  return(d);
}
/* ************************************************************************* */
ULONG UpPow2(n)
ULONG n;
{
  ULONG i,spow2;
  for(i=0,spow2=n;spow2>0;i++) spow2 = spow2>>1;
    if ((1<<(i-1)) == n)
      spow2 = n;
    else
      spow2 = 1<<i;
  return(spow2);
}
/* *********************************************************************** */
AddCommon()
{
  fprintf(orchfile,"iinstr = (int(p1)-1)*%d\n",MOT2);
  fprintf(orchfile,"inote = cpspch(p4)\n");
  fprintf(orchfile,"ioct = int(octpch(p4))\n");
  fprintf(orchfile,"io = (6+ictoct)-ioct\n");
  fprintf(orchfile,"\n");
  fprintf(orchfile,"is = (io<1 ? 1 : io)\n");
  fprintf(orchfile,"is = (is>ictoct ? ictoct : is)\n");
  fprintf(orchfile,"ipow = exp((is-1)*log(2)) ; pow(2,is-1)\n");
  fprintf(orchfile,"isn = iinstr+is\n");
  fprintf(orchfile,"\n");
  fprintf(orchfile,"iref = cpspch(ioct)\n");
  fprintf(orchfile,"in = inote/iref\n");
  fprintf(orchfile,"\n");
  fprintf(orchfile,"isampleps = isps\n");
  fprintf(orchfile,"irspc = ispc*ipow\n");
  fprintf(orchfile,"\n");
  fprintf(orchfile,"if irspc = 0 igoto nospc\n");
  fprintf(orchfile,"isampleps = irspc*inote\n");
  fprintf(orchfile,"in = 1\n");
  fprintf(orchfile,"nospc:\n");
  fprintf(orchfile,"\n");
  fprintf(orchfile,"ironeshot = ioneshot*ipow\n");
  fprintf(orchfile,"irrepeat = irepeat*ipow\n");
  fprintf(orchfile,"ionedur = ironeshot/isampleps\n");
  fprintf(orchfile,"irepdur = irrepeat/isampleps\n");
  fprintf(orchfile,"\n");
  fprintf(orchfile,"iamp = givol*ivol*p5\n");
  fprintf(orchfile,"aone    line 0,ionedur,ironeshot\n");
  fprintf(orchfile,"aout    tablei aone*in,isn,0,0,0\n");
  fprintf(orchfile,"        out     aout*iamp\n");
  fprintf(orchfile,"if irrepeat = 0 goto norepeat\n");
  fprintf(orchfile,"arep    line ionedur,irepdur,irrepeat\n");
  fprintf(orchfile,"arout   tablei arep*in,isn+%d,0,0,1\n",MAX_OCT);
  fprintf(orchfile,"        out     arout*iamp\n");
  fprintf(orchfile,"norepeat:\n");
  fprintf(orchfile,"endin\n\n");
}
/* ************************************************************************* */
/* ************************************************************************* */
LOCAL ULONG spercycle[MAX_OCT];/* # samples/cycle in high octave, else 0 */
LOCAL UWORD spersec[MAX_OCT];  /* data sampling rate */
LOCAL Fixed volume;      /* playback nominal volume from 0 to Unity */
/* ************************************************************************* */
/* Message strings for IFFP codes. */ 
LOCAL char MsgOkay[]        = { "(IFF_OKAY) No FORM 8SVX in the file." }; 
LOCAL char MsgEndMark[]     = { "(END_MARK) How did you get this message?" }; 
LOCAL char MsgDone[]        = { "(IFF_DONE) All done."}; 
LOCAL char MsgDos[]         = { "(DOS_ERROR) The DOS returned an error." }; 
LOCAL char MsgNot[]         = { "(NOT_IFF) Not an IFF file." }; 
LOCAL char MsgNoFile[]      = { "(NO_FILE) No such file found." }; 
LOCAL char MsgClientError[] = { "(CLIENT_ERROR) Read8SVX bug or insufficient RAM."}; 
LOCAL char MsgForm[]        = { "(BAD_FORM) A malformed FORM 8SVX." }; 
LOCAL char MsgShort[]       = { "(SHORT_CHUNK) A malformed FORM 8SVX." }; 
LOCAL char MsgBad[]         = { "(BAD_IFF) A mangled IFF file." }; 
 
LOCAL char *IFFPMessages[-(int)LAST_ERROR+1] = { 
    /*IFF_OKAY*/  MsgOkay, 
    /*END_MARK*/  MsgEndMark, 
    /*IFF_DONE*/  MsgDone, 
    /*DOS_ERROR*/ MsgDos, 
    /*NOT_IFF*/   MsgNot, 
    /*NO_FILE*/   MsgNoFile, 
    /*CLIENT_ERROR*/ MsgClientError, 
    /*BAD_FORM*/  MsgForm, 
    /*SHORT_CHUNK*/  MsgShort, 
    /*BAD_IFF*/   MsgBad 
    }; 
 
typedef struct { 
   ClientFrame clientFrame; 
   UBYTE foundVHDR; 
   UBYTE pad1; 
   Voice8Header sampHdr; 
   sampletype type;      /* added to handle chunk CHAN */
   } SVXFrame; 
 
 
/* NOTE: For a simple version of this program, set Fancy to 0. 
 * That'll compile a program that skips all LISTs and PROPs in the input 
 * file. It will look in CATs for FORMs 8SVX. That's suitable for most uses. 
 * 
 * For a fancy version that handles LISTs and PROPs, set Fancy to 1.
 */
#define Fancy  0

/* ************************************************************************* */
/* some local variables */
LOCAL BYTE *buf; 
LOCAL int tszBuf; 
LOCAL int szBuf; 
/* ************************************************************************* */

/** DoSomethingWithSample() **********************************************
 * 
 * Interface to Amiga sound driver. 
 * 
 *************************************************************************/
IFFP DoSomethingWithSample(sframe) SVXFrame *sframe; {
    Voice8Header *sampHdr;  
    BYTE *t;
    WORD o;
    sampHdr = &sframe->sampHdr;

/*
    printf("\noneShotHiSamples=%ld", sampHdr->oneShotHiSamples); 
    printf("\nrepeatHiSamples=%ld", sampHdr->repeatHiSamples); 
    printf("\nsamplesPerHiCycle=%ld", sampHdr->samplesPerHiCycle); 
    printf("\nsamplesPerSec=%ld", sampHdr->samplesPerSec);
    printf("\nctOctave=%ld", sampHdr->ctOctave); 
    printf("\nsCompression=%ld", sampHdr->sCompression); 
    printf("\nvolume=0x%lx", sampHdr->volume); 
    printf("\nTYPE=\n%ld", sframe->type);
*/ 

    /* Decompress, if needed. */ 
    if (sampHdr->sCompression) { 
        t = (BYTE *)calloc(tszBuf<<1,1); 
        DUnpack(buf, tszBuf, t); 
        free(buf); 
        buf = t; 
        tszBuf <<= 1; 
        };

    oneshot[0] = sampHdr->oneShotHiSamples;
    repeat[0] = sampHdr->repeatHiSamples;
    spercycle[0] = sampHdr->samplesPerHiCycle;
    spersec[0] = sampHdr->samplesPerSec;
    ctoct=sampHdr->ctOctave;
    if (ctoct > MAX_OCT) {
      printf("warning : too many octaves , keeping the %d first\n",MAX_OCT);
      ctoct = MAX_OCT;
    }
    volume = sampHdr->volume;
    for (o=0;o<ctoct-1;o++) {
      oneshot[o+1] = oneshot[o]*2;
      repeat[o+1] = repeat[o]*2;
      spercycle[o+1] = spercycle[o]*2;
      spersec[o+1] = spersec[o]*2;
    }
    if (tszBuf<((oneshot[0]+repeat[0])*((1<<ctoct)-1))) {
      printf("error not enough bytes in BODY\n");
      return(SHORT_CHUNK);
    }
    return(IFF_DONE);
} 
 
/** ReadBODY() ***********************************************************
 * 
 * Read a BODY into RAM.  
 * 
 *************************************************************************/
IFFP ReadBODY(context)  GroupContext *context;  { 
    IFFP iffp; 
 
    tszBuf  = ChunkMoreBytes(context); 
    buf  = (BYTE *)calloc(tszBuf,1); 
    if (buf == NULL) 
        iffp = CLIENT_ERROR; 
    else 
        iffp = IFFReadBytes(context, (BYTE *)buf, tszBuf); 
    CheckIFFP(); 
    } 
 
/** GetFo8SVX() **********************************************************
 * 
 * Called via ReadSample to handle every FORM encountered in an IFF file. 
 * Reads FORMs 8SVX and skips all others. 
 * Inside a FORM 8SVX, it reads BODY. It complains if it 
 * doesn't find an VHDR before the BODY. 
 * 
 * [TBD] We could read and print out any NAME and "(c) " chunks. 
 * 
 *************************************************************************/
IFFP GetFo8SVX(parent)  GroupContext *parent;  { 
   /*compilerBug register*/ IFFP iffp; 
   GroupContext formContext; 
   SVXFrame smusFrame;          /* only used for non-clientFrame fields.*/ 

   if (parent->subtype != ID_8SVX) 
      return(IFF_OKAY); /* just continue scaning the file */ 
 
   smusFrame = *(SVXFrame *)parent->clientFrame; 
   iffp = OpenRGroup(parent, &formContext); 
   CheckIFFP(); 
 
   do switch (iffp = GetFChunkHdr(&formContext)) { 
      case ID_VHDR: { 
        smusFrame.foundVHDR = TRUE; 
        iffp = GetVHDR(&formContext, &smusFrame.sampHdr); 
        break; } 
      case ID_CHAN: { 
        if (!smusFrame.foundVHDR) 
            iffp = BAD_FORM;            /* Need an VHDR chunk first! */ 
        else iffp = GetCHAN(&formContext,&smusFrame.type);
        break; } 
      case ID_BODY: { 
        if (!smusFrame.foundVHDR) 
            iffp = BAD_FORM;            /* Need an VHDR chunk first! */ 
        else iffp = ReadBODY(&formContext); 
        break; } 
      case END_MARK: { 
        if (!smusFrame.foundVHDR) 
            iffp = BAD_FORM; 
        else 
            iffp = IFF_DONE; 
        break; } 
      } while (iffp >= IFF_OKAY);  /* loop if valid ID of ignored chunk or a 
                                * subroutine returned IFF_OKAY (no errors).*/ 
 
    if (iffp != IFF_DONE)  return(iffp); 
 
    /* If we get this far, there were no errors. */ 
    CloseRGroup(&formContext); 
    iffp = DoSomethingWithSample(&smusFrame); 
    /* free(buf); */
    return(iffp); 
    } 
 
/** Notes on extending GetFo8SVX ****************************************
 * 
 * To read more kinds of chunks, just add clauses to the switch statement. 
 * To read more kinds of property chunks (like NAME) add clauses to 
 * the switch statement in GetPr8SVX, too. 
 * 
 ************************************************************************/
 
/** GetPr8SVX() *********************************************************
 * 
 * Called via ReadSample to handle every PROP encountered in an IFF file. 
 * Reads PROPs 8SVX and skips all others. 
 * 
 *************************************************************************/
#if Fancy 
IFFP GetPr8SVX(parent)  GroupContext *parent;  { 
   /*compilerBug register*/ IFFP iffp; 
   GroupContext propContext; 
   SVXFrame *svxFrame = (SVXFrame *)parent->clientFrame; /* Subclass */ 
 
   if (parent->subtype != ID_8SVX) 
      return(IFF_OKAY); /* just continue scaning the file */ 
 
   iffp = OpenRGroup(parent, &propContext); 
   CheckIFFP(); 
 
   do switch (iffp = GetPChunkHdr(&propContext)) { 
      case ID_VHDR: { 
        svxFrame->foundVHDR = TRUE; 
        iffp = GetVHDR(&propContext, &svxFrame->sampHdr); 
        break; } 
      } while (iffp >= IFF_OKAY);  /* loop if valid ID of ignored chunk or a 
                              * subroutine returned IFF_OKAY (no errors).*/ 
 
   CloseRGroup(&propContext); 
   return(iffp == END_MARK ? IFF_OKAY : iffp); 
   } 
#endif 
 
/** GetLi8SVX() **********************************************************
 * 
 * Called via ReadSample to handle every LIST encountered in an IFF file. 
 * 
 *************************************************************************/
#if Fancy 
IFFP GetLi8SVX(parent)  GroupContext *parent;  { 
    SVXFrame newFrame;  /* allocate a new Frame */ 
 
    newFrame = *(SVXFrame *)parent->clientFrame;  /* copy parent frame */ 
 
    return( ReadIList(parent, (ClientFrame *)&newFrame) ); 
    } 
#endif 
 
/** ReadSample() **********************************************************
 * 
 * Read IFF 8SVX, given a file handle open for reading.
 * 
 *************************************************************************/
IFFP ReadSample(file)  LONG file;  { 
   SVXFrame sFrame;     /* Top level "client frame".*/ 
   IFFP iffp = IFF_OKAY; 
 
#if Fancy 
   sFrame.clientFrame.getList = GetLi8SVX; 
   sFrame.clientFrame.getProp = GetPr8SVX; 
#else 
   sFrame.clientFrame.getList = SkipGroup; 
   sFrame.clientFrame.getProp = SkipGroup; 
#endif 
   sFrame.clientFrame.getForm = GetFo8SVX; 
   sFrame.clientFrame.getCat  = ReadICat ; 
 
   /* Initialize the top-level client frame's property settings to the 
    * program-wide defaults. This example just records that we haven't read 
    * any VHDR properties yet. 
    * If you want to read another property, init it's fields in sFrame. */ 
   sFrame.foundVHDR = FALSE; 
   sFrame.pad1      = 0; 
   sFrame.type      = 0;
 
   iffp = ReadIFF(file,(ClientFrame *)&sFrame); 
 
   return(iffp); 
   } 
 
/************************************************************************ */
WriteSoundIn(fout)
FILE *fout;
{

  BYTE out;
  int n0;
  short sout;

  for(n0=0;n0<szBuf;n0++) {
    out = buf[n0];
    sout = 256*(short)out;
    GWrite(fout, &sout, sizeof(short));
  }
}
/* ************************************************************************* */
void GenSound(filename,ic)
char *filename;
WORD ic;
{ 
  FILE *f;
  IFFP iffp = NO_FILE;
  char fnname[200]; 
  char fnind[200]; 
  int i;

  /* ************** */
  /* soundin part */

  if (IPath != NULL) {
    strcpy(fnname,IPath);
    strcat(fnname,"/");
    strcat(fnname,filename);
  }
  else {
    strcpy(fnname,filename);
  }
  f = GOpen(fnname, MODE_OLDFILE); 
  if (f) 
      iffp = ReadSample(f); 
  GClose(f); 
  if (iffp!=IFF_DONE) {
     printf("%s : %s\n", fnname,IFFPMessages[-iffp]); 
    return;
  }
  printf("\nconverting '%s' :\n",BaseName(filename));
  for(i=0;i<ctoct;i++) {
    strcpy(fnname,"soundin.");
    sprintf(fnind,"%d",i+1+ic*MOT2);
    strcat(fnname,fnind);
    f = GOpen(fnname,MODE_NEWFILE);
    if (f) {
        szBuf = oneshot[i];
        WriteSoundIn(f,i);
        buf += oneshot[i];
        GClose(f);
        printf("one octave %d converted and written to %s\n",i+1,fnname);
    }
    else { 
        printf("problem with file %s\n",fnname);
	  exit();
    }
    if (repeat[i] != 0) {
      strcpy(fnname,"soundin.");
      sprintf(fnind,"%d",i+MAX_OCT+1+ic*MOT2);
      strcat(fnname,fnind);
      f = GOpen(fnname,MODE_NEWFILE);
      if (f) {
          szBuf = repeat[i];
          WriteSoundIn(f,i);
          GClose(f);
          buf += repeat[i];
          printf("rep octave %d converted and written to %s\n",i+1,fnname);
      }
      else {
          printf("problem with file %s\n",fnname);
            exit();
      }
    }

  }
  /* ************** */
  /* orch file part */
  fprintf(orchfile,";---------------------------------------------------------------\n");
  fprintf(orchfile,"instr %d; %s\n",ic+1,BaseName(filename));
  fprintf(orchfile,";----- tables that should be in the score file -----\n");
  for(i=0;i<ctoct;i++) {
    fprintf(orchfile,"; f%d 0 %d 1 %d 0\n",ic*MOT2+i+1,UpPow2(oneshot[i]),ic*MOT2+i+1);
    if (repeat[i]!=0)
      fprintf(orchfile,"; f%d 0 %d 1 %d 0\n",ic*MOT2+i+MAX_OCT+1,UpPow2(repeat[i]),ic*MOT2+i+MAX_OCT+1);
  }
  fprintf(orchfile,";---------------------------------------------------\n");
  fprintf(orchfile,"ictoct   = %ld ;num of octaves\n",ctoct);
  fprintf(orchfile,"ioneshot = %ld ; # samples in the high octave 1-shot part\n",oneshot[0]);
  fprintf(orchfile,"irepeat  = %ld ; # samples in the high octave repeat part\n",repeat[0]);
  fprintf(orchfile,"ispc     = %ld ; # samples/cycle in high octave, else 0\n",spercycle[0]);
  fprintf(orchfile,"isps     = %ld ; data sampling rate \n",spersec[0]);
  fprintf(orchfile,"ivol     = %g ; playback volume \n",((float)Unity)/((float)volume));
  AddCommon();

}
/* ************************************************************************* */
GenSoundOpen(outfn)
char *outfn;
{
  char orchfn[300];

  strcpy(orchfn,outfn);
  strcat(orchfn,".orc");
  orchfile = fopen(orchfn,"w");  
  if (orchfile==NULL) {
     printf("could'nt open file %s\n",orchfn);
     exit();
  }
  printf("\nORCHESTRA FILE IS %s\n",orchfn);
  fprintf(orchfile,"; Orchestra File generated by Smus2Sco\n\n");
  fprintf(orchfile,"sr=8192 \n");
  fprintf(orchfile,"kr=512 \n");
  fprintf(orchfile,"ksmps=16 \n\n");
  fprintf(orchfile,"givol	init	30 ;global volume\n\n");
}
/* ************************************************************************* */
GenSoundClose()
{
  if (orchfile==NULL) {
     printf("error with bas fclose argument (in GenSoundClose)\n");
     exit();
  }
  fprintf(orchfile,";---------------------------------------------------------------\n");
  fclose(orchfile);
}
/* ************************************************************************* */
